#include <graphics.h>
#include <alloc.h>
#include <string.h>
#include <stdlib.h>
#include "graphic.h"
#include "standard.h"
#include "keyboard.h"
#include "path.h"
#include "actor.h"
#include "message.h"

#define DELAY_TIME_MOVE 3


ACTOR2::ACTOR2( BYTE Type, int row, int col, PathAPP *MP ):
ACTOR( 3, MP )
{
	RanDistPos( row, col, 10 );
	CurRow = RanRow;
	CurCol = RanCol;
	Process = 0;
	ProcessMask = Map->GetDirecMap( CurRow, CurCol );
	ProcessMask |= 0X10;
	DelayTimeMove = DELAY_TIME_MOVE;
	ActorMove = 1;
	DeltaRadius = 5;
	MajorType = Type;
	if( MajorType )
		GoodRandom();
	else
		BadRandom();
}

void ACTOR2::Run( void )
{
	if( Process )
		RunProcess();    	// Work follow process
	else
		RanProcess();     	// Random process
	CalCurPos();
}

void ACTOR2::SetNewPosition( int row, int col, int radius )
{
	RanDistPos( row, col, radius );
	CurRow = RanRow;
	CurCol = RanCol;
	CntStepWalk = Process = 0;
	ProcessMask = Map->GetDirecMap( CurRow, CurCol );
	ProcessMask |= 0X10;
	if( MajorType )
		GoodRandom();
	else
		BadRandom();
}

void ACTOR2::RunProcess( void )
{
	DWORD t = TIME;
	int Direc;

	if( t - TimeLastChg >= DelayTimeMove )
	{
		switch( Process ) {
		case 0X10 : // Stay
			Process = 0;
			break;

		case WALK_L : // Walk Left
			ActorMove = 1;
			CntStepWalk++;
			RemoveCol = CurCol - 1;

			if( CntStepWalk >= FullStepWalk )
			{
				RemoveCol++;
				CurCol--;
				CntStepWalk = 0;
				Direc = Map->GetDirecMap( CurRow, CurCol );
				ProcessMask = Direc;

				if( Direc == 0X05 )
					break;   			// Walk continue
				if( Direc == 0X04 )
				{
					Process = 0X04;    	// Opposite walk
					break;
				}
				// Branch path
				Process = 0;
				ProcessMask &= 0X1B; 	// Remove right direction mask
			}
			break;

		case WALK_R : // Walk Right
			ActorMove = 1;
			CntStepWalk++;
			RemoveCol = CurCol + 1;

			if( CntStepWalk >= FullStepWalk )
			{
				RemoveCol--;
				CurCol++;
				CntStepWalk = 0;
				Direc = ACTOR::Map->GetDirecMap( CurRow, CurCol );
				ProcessMask = Direc;

				if( Direc == 0X05 )
					break;   			// Walk continue
				if( Direc == 0X01 )
				{
					Process = 0X01;    	// Opposite walk
					break;
				}
				// Branch path
				Process = 0;
				ProcessMask &= 0X1E; 	// Remove left direction mask
			}
			break;

		case WALK_Up : // Walk Up
			ActorMove = 1;
			CntStepWalk++;
			RemoveRow = CurRow - 1;

			if( CntStepWalk >= FullStepWalk )
			{
				RemoveRow++;
				CurRow--;
				CntStepWalk = 0;
				Direc = ACTOR::Map->GetDirecMap( CurRow, CurCol );
				ProcessMask = Direc;

				if( Direc == 0X0A )
					break;   			// Walk continue
				if( Direc == 0X08 )
				{
					Process = 0X08;    	// Opposite walk
					break;
				}
				// Branch path
				Process = 0;
				ProcessMask &= 0X17; 	// Remove down direction mask
			}
			break;

		case WALK_Dn : // Walk Down
			ActorMove = 1;
			CntStepWalk++;
			RemoveRow = CurRow + 1;

			if( CntStepWalk >= FullStepWalk )
			{
				RemoveRow--;
				CurRow++;
				CntStepWalk = 0;
				Direc = ACTOR::Map->GetDirecMap( CurRow, CurCol );
				ProcessMask = Direc;

				if( Direc == 0X0A )
					break;   			// Walk continue
				if( Direc == 0X02 )
				{
					Process = 0X02;    	// Opposite walk
					break;
				}
				// Branch path
				Process = 0;
				ProcessMask &= 0X1D; 	// Remove up direction mask
			}
			break;
		}
		TimeLastChg = t;
		Walking = Process;
		DrawActor = ActorMove;
	}
}


// Random next process
void ACTOR2::RanProcess( void )
{
	do {
		Process = 0X01 << random(5);
		Process &= ProcessMask;
	} while( !Process );
	Walking = Process & 0X0F;
	RemoveRow = CurRow;
	RemoveCol = CurCol;
}

void ACTOR2::GoodRandom( void )
{
	int PropRan[] = { 4, 10, 18, 28, 43, 63, 83, 103, 123, 153,
					  183, 195, 235, 250 };
	int DelRad[] = { 35, 25, 23, 20, 18, 18, 15, 12, 12,
					 10, 10, 15, 8, 7 };
	MinorType = random( 250 );
	int Type;
	for( Type = 0 ; Type < NUM_GOOD_ACTOR ; Type++ )
		if( MinorType < PropRan[Type] )
			break;
	MinorType = Type;
	DeltaRadius = DelRad[MinorType];
//	MinorType += ST_GOOD_ACTOR;
}

void ACTOR2::BadRandom( void )
{
	const int RadProp[] = { 1, 0, 1, 0, 0, 0, 2, 0, 0, 1, 0, 1, 0, 0, 0 };
	MinorType = ST_BAD_ACTOR + random( (int)NUM_BAD_ACTOR );
	DeltaRadius = 1;
	for( register int i = 0 ; i < 30 ; i++ )
		DeltaRadius += RadProp[random(15)];
}
