#include <graphics.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <alloc.h>
#include <dos.h>
#include <string.h>
#include <ctype.h>
#include "graphic.h"

unsigned char far *ScreenLinePointers[480];
int _ScrLeft = 0, _ScrRight = 639, _ScrTop = 0, _ScrBottom = 479;
#define TAB_SKIP 16

typedef union {
	unsigned int Word;
	struct {
		unsigned char low, high;
		} Byte;
	}  WordUnion;


/*
///////////////////////////////////////////////////////////
// StartGraphics()
//
// This function initializes Borland's graphics driver
// for the high-resolution VGA screen.
///////////////////////////////////////////////////////////
*/
void StartGraphics(void)
{
	int gdriver = VGA, gmode = VGAHI, errorcode;

	errorcode = registerbgidriver( EGAVGA_driver );
	if (errorcode < 0)
	{
		printf("Graphics not initialized: %d\n",errorcode);
	  abort();
	}

	initgraph(&gdriver, &gmode, "");
	if ((errorcode = graphresult()) != grOk)
	{
		printf(" Graphics not initialized: %d\n",errorcode);
		abort();
	}

	for( register line = 0 ; line < 480 ; line++ )
		ScreenLinePointers[line] =
			(unsigned char far *) MK_FP( 0XA000, 80*line );
}

// Variable for gprintf & Gprintf Module
char PrintString[512];

/*
int gprintf( int x, int y, char *fmt, ... )
{
	va_list argptr;
	int cnt;
	va_start( argptr, fmt );
	cnt = vsprintf( PrintString, fmt, argptr );
	outtextxy( x, y, PrintString );
	va_end( argptr );
	return cnt;
}
*/

int Gprintf( int x, int y, char *fmt, ... )
{
	va_list argptr;
	char OutChar[2] = " ";
	int cnt, CntLine = 0;
	va_start( argptr, fmt );
	cnt = vsprintf( PrintString, fmt, argptr );
	va_end( argptr );
	int TextColor, NormalColor = getcolor();
	int height = 3*textheight( "A" )/2;

	char *StrPtr = PrintString;

	moveto( x, y );

	while( *StrPtr )
	{
		OutChar[0] = *StrPtr++;
		switch( OutChar[0] ) {
		case '$' :
			if( isxdigit( *StrPtr ) )
			{
				if( isdigit( *StrPtr ) )
					TextColor = (*StrPtr) - '0';
				else
					TextColor = (*StrPtr) - 'A' + 10;
				setcolor( TextColor );
				StrPtr++;
				continue;
			}
			if( *StrPtr == '$' )	// Double $
				StrPtr++;
			else       				// Single $
			{
				TextColor = NormalColor;
				continue;
			}
			break;

		case '\n' :
			CntLine++;
			moveto( x, y+CntLine*height );
			continue;

		case '\t' :
			moverel( TAB_SKIP, 0 );
			continue;
		}
		outtext( OutChar );
	}
	return cnt;
}

int TextWidth( char *fmt, ... )
{
	va_list argptr;
	int CntLen = 0, MaxLen = 0;
	va_start( argptr, fmt );
	vsprintf( PrintString, fmt, argptr );
	va_end( argptr );

	char *StrPtr = PrintString;

	while( *StrPtr )
	{
		char ch = *StrPtr++;
		switch( ch ) {
		case '$' :
			if( isxdigit( *StrPtr ) )
			{
				StrPtr++;
				continue;
			}
			if( *StrPtr == '$' )    // Double $ (Write $)
			{
				StrPtr++;
				CntLen++;
			}
			continue;

		case '\n' :
			if( CntLen > MaxLen )
				MaxLen = CntLen;
			CntLen = 0;
			continue;
		}
		CntLen++;
	}
	if( CntLen > MaxLen )
		MaxLen = CntLen;
	return MaxLen*textwidth("A");
}

int TextHeight( char *fmt, ... )
{
	int CntLine = 1;
	va_list argptr;
	va_start( argptr, fmt );
	vsprintf( PrintString, fmt, argptr );
	va_end( argptr );
	int height = textheight( "A" );
	char *StrPtr = PrintString;

	do {
		StrPtr = strchr( StrPtr, '\n' );
		if( StrPtr )
		{
			CntLine++;
			StrPtr++;
		}
		else
			break;
	} while( 1 );

	return ( 3*CntLine*height - height ) / 2;
}

void SetPlane( int plane )
{
	outp( 0X03C4, 2 );
	outp( 0X03C5, plane );
}

void SetBltBit( int bltbit )
{
	const unsigned char BitBlt[4] = { 0, 0X18, 0X10, 0X08 };
	outp( 0X03CE, 3 );
	outp( 0X03CF, BitBlt[bltbit] );
}


void SetBitPlane( int plane, int line, int x, int w,
				  MapPtr LinePtr, int bitblt )
{
	int ScreenLineOffset, ScreenLineMask;
	int LineOffset, LineMask;
	ScreenLineMask = LineMask = 0X80;
	int CntW;
	const unsigned char BitBlt[4] = { 0, 0X18, 0X10, 0X08 };

	w = min( w, _ScrRight - x );
	if( x >= _ScrLeft )
	{
		ScreenLineOffset = x/8;
		LineOffset = 0;
		ScreenLineMask >>= x%8;
		CntW = 0;
	}
	else
	{
		x = _ScrLeft - x;
		ScreenLineOffset = 0;
		LineOffset = x/8;
		LineMask >>= x%8;
		CntW = x;
	}

	MapPtr ScreenLinePtr = ScreenLinePointers[line] + ScreenLineOffset;
	LinePtr += LineOffset;

	if( bitblt >= COPY_PUT && bitblt <= AND_PUT )
	{
		outp( 0X03C4, 2 );
		outp( 0X03C5, plane );
		outp( 0X03CE, 3 );
		outp( 0X03CF, BitBlt[bitblt] );

		for( ; CntW <= w ; CntW++ )
		{
			if( *LinePtr & LineMask )
				*ScreenLinePtr |= ScreenLineMask;
			else
				*ScreenLinePtr &= ScreenLineMask;

			if( ScreenLineMask & 0X01 )
			{
				ScreenLineMask = 0X80;
				ScreenLinePtr++;
			}
			else
				ScreenLineMask >>= 1;

			if( LineMask & 0X01 )
			{
				LineMask = 0X80;
				LinePtr++;
			}
			else
				LineMask >>= 1;
		}
		outp( 0X03C4, 2 );
		outp( 0X03C5, 0 );
	}
}


void PutPicture( int x, int y, void far *bitmap, int bitblt )
{
	ImaInfo Info;
	ImageInfo( &Info, bitmap );
	int CntLine, MaxLine;
	MapPtr Plane = (MapPtr) bitmap + 4;

	MaxLine = min( Info.MaxY, _ScrBottom - y );
	if( y >= _ScrTop )
		CntLine = 0;
	else
		CntLine = _ScrTop - y;


	for( ; CntLine <= MaxLine ; CntLine++ )
	{
		SetBitPlane( 1, y+CntLine, x, Info.MaxX, Plane, bitblt );
		Plane += Info.BytePerPlane;
		SetBitPlane( 2, y+CntLine, x, Info.MaxX, Plane, bitblt );
		Plane += Info.BytePerPlane;
		SetBitPlane( 4, y+CntLine, x, Info.MaxX, Plane, bitblt );
		Plane += Info.BytePerPlane;
		SetBitPlane( 8, y+CntLine, x, Info.MaxX, Plane, bitblt );
		Plane += Info.BytePerPlane;
	}
}

void PutImage( int left, int top, void far *bitmap, int option )
{
	ImaInfo Info;
	ImageInfo( &Info, bitmap );
	int right = Info.MaxX;
	int bottom = Info.MaxY;

	unsigned int Color;
	unsigned char ColorMask;
	int PixelX, PixelY, PixelC;

	MapPtr Plane = (MapPtr)bitmap + 4;
	MapPtr BitPlane[4];
	unsigned char BitMask;
	int PlaneOffset;
	unsigned char BitMaskX = 0X80;

	if( left < 0 )
	{
		PixelX = 0;
		left = -left;
		BitMaskX >>= left%8;
	}
	else
	{
		PixelX = left;
		left = 0;
	}
	int LeftMark = left;
	int PixelXMark = PixelX;

	PlaneOffset = left / 8;
	if( top < 0 )
	{
		PixelY = 0;
		top = -top;
	}
	else
	{
		PixelY = top;
		top = 0;
	}

	for( ; top <= bottom && PixelY < 480 ; top++, PixelY++ )
	{
		BitPlane[0] = Plane + top*Info.BytePerLine + PlaneOffset;
		BitPlane[1] = BitPlane[0] + Info.BytePerPlane;
		BitPlane[2] = BitPlane[1] + Info.BytePerPlane;
		BitPlane[3] = BitPlane[2] + Info.BytePerPlane;
		BitMask = BitMaskX;

		for( left = LeftMark, PixelX = PixelXMark ;
			 left <= right && PixelX < 640 ;
			 left++, PixelX++ )
		{
			PixelC = getpixel( PixelX, PixelY );
			Color = 0;
			for( register i = 0, ColorMask = 0X08 ; i < 4 ;
				 i++, ColorMask >>= 1 )
			{
				if( *BitPlane[i] & BitMask )
					Color |= ColorMask;
			}

			switch( option ) {
			case COPY_PUT :
				putpixel( PixelX, PixelY, Color );
				break;
			case XOR_PUT :
				putpixel( PixelX, PixelY, Color^PixelC );
				break;
			case OR_PUT :
				putpixel( PixelX, PixelY, Color|PixelC );
				break;
			case AND_PUT :
				putpixel( PixelX, PixelY, Color&PixelC );
				break;
			case NOT_PUT :
				putpixel( PixelX, PixelY, Color^0XFF );
			case NOBLACK :
				if( Color )
					putpixel( PixelX, PixelY, Color );
				break;
			}

			if( BitMask & 0X01 )
			{
				BitMask = 0X80;
				BitPlane[0]++;
				BitPlane[1]++;
				BitPlane[2]++;
				BitPlane[3]++;
			}
			else
				BitMask >>= 1;
		}
	}
}

void ImageInfo( ImaInfo *Info, void far *bitmap )
{
	int far *IntPtr = (int far *) bitmap;
	Info->MaxX = *IntPtr++;
	Info->MaxY = *IntPtr;
	Info->BytePerPlane = Info->MaxX/8 + 1;
	Info->BytePerLine = 4*Info->BytePerPlane;
}

void ImageInfo( ImaInfo *Info, void far *bitmap, int x, int y )
{
	ImageInfo( Info, bitmap );
	Info->Offset = y*Info->BytePerLine + x/8;
	Info->Mask = 0X80 >> x%8;
}


void ImaMoveLeft( MapPtr Image, int DeltaX )
{
	ImaInfo Info;
	ImageInfo( &Info, Image );
	WordUnion Mask, EndMask;
	int RoL = DeltaX % 8;

	int FullMaskMove = Info.MaxX - DeltaX + 1;
					// the number of pixel that will be removed
	EndMask.Word = 0XFF00;
	EndMask.Word >>= FullMaskMove % 8;
	EndMask.Byte.high ^= 0XFF;
	FullMaskMove /= 8;    // FullMaskMove = ( Info.MaxX - DeltaX + 1 ) / 8;

	int j, MaxJ1 = Info.MaxX / 8;

	MapPtr BitPlane = (MapPtr) Image + 4;
	MapPtr Scr, Des;
	int Offset = DeltaX / 8;

	for( int CntY = 0 ; CntY <= Info.MaxY ; CntY++ )
	{
		for( int p = 0 ; p < 4 ; p++, BitPlane += Info.BytePerPlane )
		{
			Des = BitPlane;
			Scr = Des + Offset;
			for( j = 0 ; j < FullMaskMove ; j++ )
			{
				Mask.Byte.high = *Scr++;
				Mask.Byte.low = *Scr;
				Mask.Word <<= RoL;
				*Des++ = Mask.Byte.high;
			}
			if( EndMask.Byte.high )
			{
				Mask.Byte.high = *Scr++;
				Mask.Byte.low = *Scr;
				Mask.Word <<= RoL;
				*Des++ = Mask.Byte.high & EndMask.Byte.high;
				j++;
			}
			for( ; j <= MaxJ1 ; j++ )
				*Des++ = 0;
		}
	}
}

void ImaMoveRight( MapPtr Image, int DeltaX )
{
	ImaInfo Info;
	ImageInfo( &Info, Image );

	int FullMaskMove = Info.MaxX - DeltaX + 1;
			// the number of pixel that will be removed
	FullMaskMove /= 8;    // FullMaskMove = ( Info.MaxX - DeltaX + 1 ) / 8;

	WordUnion Mask, EndMask;
	int RoR =  DeltaX % 8;

	EndMask.Word = 0X00FF;
	EndMask.Word >>= DeltaX % 8;

	int j, MaxJ1 = Info.MaxX / 8;

	MapPtr BitPlane = (MapPtr) Image + 4 + Info.MaxX/8;
	MapPtr Scr, Des;
	int Offset = DeltaX / 8;

	for( int CntY = 0 ; CntY <= Info.MaxY ; CntY++ )
	{
		for( int p = 0 ; p < 4 ; p++, BitPlane += Info.BytePerPlane )
		{
			Des = BitPlane;
			Scr = Des - Offset;
			for( j = 0 ; j < FullMaskMove ; j++ )
			{
				Mask.Byte.low = *Scr--;
				Mask.Byte.high = *Scr;
				Mask.Word >>= RoR;
				*Des-- = Mask.Byte.low;
			}
			if( EndMask.Byte.low )
			{
				Mask.Byte.low = *Scr--;
				Mask.Byte.high = *Scr;
				Mask.Word >>= RoR;
				*Des-- = Mask.Byte.low & EndMask.Byte.low;
				j++;
			}
			for( ; j <= MaxJ1 ; j++ )
				*Des-- = 0;
		}
	}
}

void ImaMoveUp( MapPtr Image, int DeltaY )
{
	ImaInfo Info;
	ImageInfo( &Info, Image );
	MapPtr DesPtr = Image + 4;
	MapPtr SrcPtr = DesPtr + DeltaY*Info.BytePerLine;
	int MaxCntY = Info.MaxY - DeltaY;

	for( int CntY = 0 ; CntY <= MaxCntY ; CntY++ )
	{
		memcpy( (void *)DesPtr, (void *)SrcPtr, Info.BytePerLine );
		DesPtr += Info.BytePerLine;
		SrcPtr += Info.BytePerLine;
	}
	memset( (void *)DesPtr, 0, DeltaY*Info.BytePerLine );
}

void ImaMoveDn( MapPtr Image, int DeltaY )
{
	ImaInfo Info;
	ImageInfo( &Info, Image );
	MapPtr DesPtr = Image + 4;
	DesPtr += Info.MaxY * Info.BytePerLine;
	MapPtr SrcPtr = DesPtr - DeltaY*Info.BytePerLine;
	int MaxCntY = Info.MaxY - DeltaY;

	for( int CntY = 0 ; CntY <= MaxCntY ; CntY++ )
	{
		memcpy( (void *)DesPtr, (void *)SrcPtr, Info.BytePerLine );
		DesPtr -= Info.BytePerLine;
		SrcPtr -= Info.BytePerLine;
	}
	DesPtr = Image + 4;
	memset( (void *)DesPtr, 0, DeltaY*Info.BytePerLine );
}

void ImaLine( MapPtr Image, int x1, int y1, int x2, int y2, int color )
{

	int DeltaX = x2 - x1;
	int AbsDelX = abs( DeltaX );
	int AbsDelX2 = AbsDelX / 2;
	DeltaX = ( DeltaX >= 0) ? 1 : -1;

	int DeltaY = y2 - y1;
	int AbsDelY = abs( DeltaY );
	int AbsDelY2 = AbsDelY / 2;
	DeltaY = ( DeltaY >= 0 ) ? 1 : -1;

	int XInc, YInc;
	int Increase = max( AbsDelX, AbsDelY );

	ImaInfo Info;
	ImageInfo( &Info, Image );
	MapPtr BitPlane[4];
	unsigned char BitMask, ColorMask;
	MapPtr Plane = (MapPtr) Image + 4;
	int i;

	XInc = YInc = 0;
	for( int Inc = 0 ; Inc <= Increase ; Inc++ )
	{
		if( x1 >= 0 && y1 >= 0 && x1 <= Info.MaxX && y1 <= Info.MaxY)
		{
			BitPlane[0] = Plane + y1*Info.BytePerLine + x1/8;
			BitPlane[1] = BitPlane[0] + Info.BytePerPlane;
			BitPlane[2] = BitPlane[1] + Info.BytePerPlane;
			BitPlane[3] = BitPlane[2] + Info.BytePerPlane;
			BitMask = 0X80;
			BitMask >>= x1%8;
			for( ColorMask = 0X08, i = 0 ; i < 4 ; i++, ColorMask >>= 1 )
			{
				if( ColorMask & color )
					*BitPlane[i] |= BitMask;
				else
					*BitPlane[i] &= BitMask^0XFF;
			}
		}

		XInc += AbsDelX;
		YInc += AbsDelY;
		if( XInc + AbsDelX2 >= Increase )
		{
			XInc -= Increase;
			x1 += DeltaX;
		}
		if( YInc + AbsDelY2 >= Increase )
		{
			YInc -= Increase;
			y1 += DeltaY;
		}
	}
}

void ImaBar( MapPtr Image, int x1, int y1, int x2, int y2, int color )
{
	if( y1 > y2 )
	{
		y1 ^= y2;
		y2 ^= y1;
		y1 ^= y2;
	}
	for( ; y1 <= y2 ; y1++ )
		ImaLine( Image, x1, y1, x2, y1, color );
}



void ImaFillCircle( MapPtr Image, int x, int y, int radius, int color )
{
	ImaInfo Info;
	ImageInfo( &Info, Image );

	MapPtr Plane = Image + 4;
	MapPtr Des, DesMark;

	long int DummyY, DummyY1, AbsDummyY, AbsDummyY1;
	long int RadSqrt = (long) (radius * radius);
	int VarY, MarkY;
	long int VarX2;
	int VarX;

	int FirstPixelPlot, EndPixelPlot;
	int FirstOffset, EndOffset;
	unsigned char FirstMask, EndMask;
	int CntFullMask, FullMaskByte;
	int plane;
	unsigned char CMask, CAndMask;

	int CntY;

	for( VarX = 0, VarY = radius ; VarY >= 0 ; VarY-- )
	{
		DummyY = RadSqrt - (long)VarY*VarY;
		DummyY1 = RadSqrt - (long)(VarY-1)*(VarY-1);
		for( ; VarX < radius ; VarX++ )
		{
			VarX2 = VarX+1;
			VarX2 *= VarX2;
			AbsDummyY = abs( DummyY - VarX2 );
			AbsDummyY1 = abs( DummyY1- VarX2 );
			if( AbsDummyY1 < AbsDummyY )
				break;
		}

		FirstPixelPlot = max( 0, x-VarX );
		if( FirstPixelPlot > Info.MaxX )
			continue;
		EndPixelPlot = min( Info.MaxX, x+VarX );
		if( EndPixelPlot < 0 )
			continue;

		FirstMask = 0XFF >> (FirstPixelPlot%8);
		FirstOffset = FirstPixelPlot / 8;

		EndOffset = (EndPixelPlot+1) / 8;
		EndMask = 0XFF >> ((EndPixelPlot+1)%8);
		EndMask ^= 0XFF;

		if( FirstOffset == EndOffset )
		{
			if( EndMask )
				FirstMask &= EndMask;
			FullMaskByte = EndMask = 0;
		}
		else
			FullMaskByte = EndOffset - FirstOffset - 1;


		// CntY = -1 ==> Draw Top Half of Circle
		// CntY =  1 ==> Draw Bottom Half of Circle

		for( CntY = -1 ; CntY <= 1 ; CntY += 2 )
		{
			MarkY = y - CntY * VarY;
			if( MarkY >= 0 && MarkY <= Info.MaxY )
			{
				DesMark = Plane + MarkY*Info.BytePerLine + FirstOffset;
				CMask = 0X08;
				for( plane = 0 ; plane < 4 ; plane++,
					 DesMark += Info.BytePerPlane, CMask >>= 1 )
				{
					if( CMask & color )
						CAndMask = 0XFF;
					else
						CAndMask = 0;

					Des = DesMark;

					if( CAndMask )
						*Des |= FirstMask;
					else
						*Des &= FirstMask ^ 0XFF;
					Des++;

					for( CntFullMask = 0 ; CntFullMask < FullMaskByte ;
						 CntFullMask++ )
					{
						if( CAndMask )
							*Des++ = 0XFF;
						else
							*Des++ = 0;
					}

					if( EndMask )
					{
						if( CAndMask )
							*Des |= EndMask;
						else
							*Des &= EndMask ^ 0XFF;
					}
				}
			}
		}
	}
}


int ImaWriteFile( MapPtr Image, char *filename )
{
	int far *IntPtr = (int far *)Image;
	FILE *fout = fopen( filename, "wb" );
	if( fout == NULL )
		return 0;

	int PixelX = *IntPtr++;
	int PixelY = *IntPtr;
	size_t size = imagesize( 0, 0, PixelX, PixelY );

	fputc( (size&0XFF00)>>8, fout );
	fputc( size&0X00FF, fout );
	if( fwrite( (void *)Image, 1, size, fout ) != size )
	{
		fclose( fout );
		return 0;
	}
	fclose( fout );
	return 1;
}

int ImaReadFile( MapPtr *Image, char *filename )
{
	unsigned char ch;
	MapPtr ImaPick;

	FILE *fin = fopen( filename, "rb" );
	if( fin == NULL )
		return 0;

	size_t size = fgetc( fin );
	size <<= 8;
	ch = fgetc( fin );
	size |= ch;

	ImaPick = (MapPtr) farmalloc( size );
	if( !ImaPick )
	{
		*Image = NULL;
		return 0;
	}
			 if( fread( (void *)ImaPick, 1, size, fin ) != size )
	{
		fclose( fin );
		*Image = NULL;
		return 0;
	}
	fclose( fin );
	*Image = ImaPick;
	return 1;
}


void ImaBltBit( void far *back, int x, int y, void far *fore, int bltbit )
{
	ImaInfo ForeInfo, BackInfo;
	ImageInfo( &ForeInfo, fore );
	ImageInfo( &BackInfo, back );

	WordUnion ForeMask, BackMask;
	WordUnion BackMaskInv;
	WordUnion ForeCmp, BackCmp;
	WordUnion EndMask, EndMaskInv;

	int CmpRoR, FullMaskX;

	MapPtr Fore = (MapPtr) fore + 4;
	MapPtr Back = (MapPtr) back + 4;
	MapPtr Src, Des;

	if( y >= 0 )
		Back += y*BackInfo.BytePerLine;
	else
		Fore -= y*ForeInfo.BytePerLine;

	ForeMask.Word = BackMask.Word = 0XFF00;
	EndMask.Word = 0XFFFF;
	int CntByte, FullMaskCmp;

	CmpRoR = x % 8;  // OK for rotate right ForeMask
	if( x >= 0 )
	{
		BackMask.Word >>= CmpRoR;
		EndMask.Word >>= CmpRoR;
		Back += x/8;

		FullMaskX = min( ForeInfo.MaxX+1, BackInfo.MaxX-x+1 );
		FullMaskCmp = FullMaskX / 8;
	}
	else
	{
		ForeMask.Word >>= -CmpRoR;
		EndMaskInv.Word >>= -CmpRoR;
		Fore -= x/8;

		FullMaskX = min( ForeInfo.MaxX+x+1, BackInfo.MaxX+1 );
		FullMaskCmp = FullMaskX / 8;
	}

	BackMaskInv.Word = BackMask.Word ^ 0XFFFF;

	EndMaskInv.Word = EndMask.Word;
	EndMaskInv.Word >>= (FullMaskX%8);
	EndMask.Word ^= EndMaskInv.Word;
	EndMaskInv.Word = EndMask.Word ^ 0XFFFF;


	int CntY = max( 0, -y );
	int MaxCntY = min( ForeInfo.MaxY, BackInfo.MaxY-y );
	int plane;

	switch( bltbit ) {
	case COPY_PUT :
		for( ; CntY <= MaxCntY ; CntY++ )
		{
			for( plane = 0 ; plane < 4 ; plane++,
				 Fore += ForeInfo.BytePerPlane,
				 Back += BackInfo.BytePerPlane )
			{
				Src = Fore;
				Des = Back;
				for( CntByte = 0 ; CntByte < FullMaskCmp ; CntByte++ )
				{
					ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
					ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
					ForeCmp.Word >>= CmpRoR;
					BackCmp.Byte.high = *Des;
					BackCmp.Byte.low = *(Des+1);
					BackCmp.Word =
						( BackCmp.Word & BackMaskInv.Word ) |
						( ForeCmp.Word & BackMask.Word );
					*Des++ = BackCmp.Byte.high;
					*Des = BackCmp.Byte.low;
				}

				ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
				ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
				ForeCmp.Word >>= CmpRoR;
				BackCmp.Byte.high = *Des;
				BackCmp.Byte.low = *(Des+1);
				BackCmp.Word =
					( BackCmp.Word & EndMaskInv.Word ) |
					( ForeCmp.Word & EndMask.Word );
				*Des++ = BackCmp.Byte.high;
				*Des = BackCmp.Byte.low;
			}
		}
		return;

	case XOR_PUT :
		for( ; CntY <= MaxCntY ; CntY++ )
		{
			for( plane = 0 ; plane < 4 ; plane++,
				 Fore += ForeInfo.BytePerPlane,
				 Back += BackInfo.BytePerPlane )
			{
				Src = Fore;
				Des = Back;
				for( CntByte = 0 ; CntByte < FullMaskCmp ; CntByte++ )
				{
					ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
					ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
					ForeCmp.Word >>= CmpRoR;
					BackCmp.Byte.high = *Des;
					BackCmp.Byte.low = *(Des+1);
					BackCmp.Word =
						( BackCmp.Word & BackMaskInv.Word ) |
						( (ForeCmp.Word^BackCmp.Word) & BackMask.Word );
					*Des++ = BackCmp.Byte.high;
					*Des = BackCmp.Byte.low;
				}

				ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
				ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
				ForeCmp.Word >>= CmpRoR;
				BackCmp.Byte.high = *Des;
				BackCmp.Byte.low = *(Des+1);
				BackCmp.Word =
					( BackCmp.Word & EndMaskInv.Word ) |
					( (ForeCmp.Word^BackCmp.Word) & EndMask.Word );
				*Des++ = BackCmp.Byte.high;
				*Des = BackCmp.Byte.low;
			}
		}
		return;

	case OR_PUT :
		for( ; CntY <= MaxCntY ; CntY++ )
		{
			for( plane = 0 ; plane < 4 ; plane++,
				 Fore += ForeInfo.BytePerPlane,
				 Back += BackInfo.BytePerPlane )
			{
				Src = Fore;
				Des = Back;
				for( CntByte = 0 ; CntByte < FullMaskCmp ; CntByte++ )
				{
					ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
					ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
					ForeCmp.Word >>= CmpRoR;
					BackCmp.Byte.high = *Des;
					BackCmp.Byte.low = *(Des+1);
					BackCmp.Word =
						( BackCmp.Word & BackMaskInv.Word ) |
						( (ForeCmp.Word|BackCmp.Word) & BackMask.Word );
					*Des++ = BackCmp.Byte.high;
					*Des = BackCmp.Byte.low;
				}

				ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
				ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
				ForeCmp.Word >>= CmpRoR;
				BackCmp.Byte.high = *Des;
				BackCmp.Byte.low = *(Des+1);
				BackCmp.Word =
					( BackCmp.Word & EndMaskInv.Word ) |
					( (ForeCmp.Word|BackCmp.Word) & EndMask.Word );
				*Des++ = BackCmp.Byte.high;
				*Des = BackCmp.Byte.low;
			}
		}
		return;

	case AND_PUT :
		for( ; CntY <= MaxCntY ; CntY++ )
		{
			for( plane = 0 ; plane < 4 ; plane++,
				 Fore += ForeInfo.BytePerPlane,
				 Back += BackInfo.BytePerPlane )
			{
				Src = Fore;
				Des = Back;
				for( CntByte = 0 ; CntByte < FullMaskCmp ; CntByte++ )
				{
					ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
					ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
					ForeCmp.Word >>= CmpRoR;
					BackCmp.Byte.high = *Des;
					BackCmp.Byte.low = *(Des+1);
					BackCmp.Word =
						( BackCmp.Word & BackMaskInv.Word ) |
						( (ForeCmp.Word&BackCmp.Word) & BackMask.Word );
					*Des++ = BackCmp.Byte.high;
					*Des = BackCmp.Byte.low;
				}

				ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
				ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
				ForeCmp.Word >>= CmpRoR;
				BackCmp.Byte.high = *Des;
				BackCmp.Byte.low = *(Des+1);
				BackCmp.Word =
					( BackCmp.Word & EndMaskInv.Word ) |
					( (ForeCmp.Word&BackCmp.Word) & EndMask.Word );
				*Des++ = BackCmp.Byte.high;
				*Des = BackCmp.Byte.low;
			}
		}
		return;

	case NOT_PUT :
		for( ; CntY <= MaxCntY ; CntY++ )
		{
			for( plane = 0 ; plane < 4 ; plane++,
				 Fore += ForeInfo.BytePerPlane,
				 Back += BackInfo.BytePerPlane )
			{
				Src = Fore;
				Des = Back;
				for( CntByte = 0 ; CntByte < FullMaskCmp ; CntByte++ )
				{
					ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
					ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
					ForeCmp.Word >>= CmpRoR;
					BackCmp.Byte.high = *Des;
					BackCmp.Byte.low = *(Des+1);
					BackCmp.Word =
						( BackCmp.Word & BackMaskInv.Word ) |
						( (ForeCmp.Word^0XFFFF) & BackMask.Word );
					*Des++ = BackCmp.Byte.high;
					*Des = BackCmp.Byte.low;
				}

				ForeCmp.Byte.high = *Src++ & ForeMask.Byte.high;
				ForeCmp.Byte.low = *Src & ForeMask.Byte.low;
				ForeCmp.Word >>= CmpRoR;
				BackCmp.Byte.high = *Des;
				BackCmp.Byte.low = *(Des+1);
				BackCmp.Word =
					( BackCmp.Word & EndMaskInv.Word ) |
					( (ForeCmp.Word^0XFFFF) & EndMask.Word );
				*Des++ = BackCmp.Byte.high;
				*Des = BackCmp.Byte.low;
			}
		}
		return;
	}
}
