/*
ͻ
  FILENAME : Path.CPP                                     
  COMPILER : Turbo C++ Version 1.00                       
  INCLUDE  : Path.H, Graphic.H, Standard.H                
͹
  Present by MR.Watcharop   Riangrungrote (WRR)           
             Mechanical   Engineering                     
             Chulalongkorn University                     
ͼ
*/
#include <graphics.h>
#include <stdio.h>
#include <alloc.h>
#include <stdlib.h>
#include <string.h>
#include <dir.h>
#include "standard.h"
#include "graphic.h"
#include "path.h"

#define GAME

PathAPP::PathAPP( int row, int col, BYTE PPB, MapPtr floor )
{
	Complete = 1;

	StartX = StartY = 0;
	MaxRow = row;
	MaxCol = col;

	strcpy( (char *)Designer, "" );
	PixelPerBlock = PPB;
	WallColor = RED;  // Default Color
	MaxFinal = 0;
	memset( (void *)&Final, 0, 4*MAX_FINAL );

	FloorIma = NULL;
	SizeFloor = imagesize( 1, 1, PixelPerBlock, PixelPerBlock );
	if( !SetFloorIma(floor) )
		Complete = 0;

	SizeMap = row * ((col-1)/8 + 1);
	Map = (MapPtr) farmalloc( SizeMap );
	if( Map )
		memset( (void *)Map, 0, SizeMap );
	else
		Complete = 0;

	FinalIma = (MapPtr) NULL;
	if( !ReadFinalIma() )
		Complete = 0;

}

PathAPP::PathAPP( char *filename )
{
	Complete = 1;
	FloorIma = Map = (MapPtr) NULL;
	if( !ReadFile( filename ) )
		Complete = 0;

	FinalIma = (MapPtr) NULL;
	if( !ReadFinalIma() )
		Complete = 0;

#if defined(GAME)
	if( Complete )
	{
		if( MaxFinal )
			FinalSelected = random( MaxFinal );
		else
			RandomFinal();
	}
#endif
}

BYTE PathAPP::ReadFinalIma( void )
{
	if( FinalIma )
		return 1;

	if( !ImaReadFile( &FinalIma, "Final.IMA" ) )
		return 0;

	ImaInfo Info;
	ImageInfo( &Info, FinalIma );
	int PPB = PixelPerBlock - 1;
	if( Info.MaxX != PPB || Info.MaxY != PPB )
		return 0;
	return 1;
}

BYTE PathAPP::SetFloorIma( MapPtr floor )
{
	if( !floor )
		return 0;

	ImaInfo Info;
	ImageInfo( &Info, floor );
	int PPB = PixelPerBlock - 1;

	if( Info.MaxX != PPB || Info.MaxY != PPB )
	{
		farfree( floor );
		return 0;
	}

	if( FloorIma )
		farfree( FloorIma );

	FloorIma = floor;
	return 1;
}


BYTE PathAPP::Add( int row, int col )
{
	if( !HaveFloor( row, col ) )
	{
		*(Map+PosByte) |= PosBit;
		return 1;
	}
	return 0;
}

// return 0 no change
// return 1 add final area
// return -1 remove final area
int PathAPP::ToggleFinal( int row, int col )
{
	if( !HaveFloor( row, col ) )
		return 0;

	for( int i = 0 ; i < MaxFinal ; i++ )
	{
		if( Final[i].Row == row && Final[i].Col == col )
		{
			MaxFinal--;
			Final[i].Row = Final[MaxFinal].Row;
			Final[i].Col = Final[MaxFinal].Col;
			Final[MaxFinal].Row = Final[MaxFinal].Col = 0;
			return -1;
		}
	}
	// Add final area
	if( MaxFinal < MAX_FINAL )
	{
		Final[MaxFinal].Row = row;
		Final[MaxFinal].Col = col;
		MaxFinal++;
		return 1;
	}
	return 0;
}

void PathAPP::RandomFinal( void )
{
	while( 1 )
	{
		int row = random( 40 );
		int col = random( 40 );
		if( HaveFloor( row, col ) )
		{
			FinalSelected = 0;
			MaxFinal = 1;
			Final[0].Row = row;
			Final[0].Col = col;
			return;
		}
	}
}

BYTE PathAPP::Remove( int row, int col )
{
	if( HaveFloor( row, col ) )
	{
		*(Map+PosByte) &= (PosBit^0XFF);
		return 1;
	}
	return 0;
}

void PathAPP::CalHoldPoint( int row, int col, int *x, int *y )
{
	*x = StartX + col * PixelPerBlock;
	*y = StartY + row * PixelPerBlock;
}

void PathAPP::CalXYBorder( int row, int col )
{
	BYTE PPB = PixelPerBlock - 1;
	XBorder[1] = XBorder[0] = StartX + col * PixelPerBlock;
	XBorder[1] += PPB;
	YBorder[1] = YBorder[0] = StartY + row * PixelPerBlock;
	YBorder[1] += PPB;
}

void PathAPP::DrawMap( void )
{
	int i;
	for( int j = 0 ; j < MaxRow ; j++ )
		for( i = 0 ; i < MaxCol ; i++ )
			DrawFloor( i, j );
}

int PathAPP::GetDirecMap( int row, int col )
{
	int Direc = 0;
	if( HaveFloor( row, col-1 ) )
		Direc |= 1;
	if( HaveFloor( row, col+1 ) )
		Direc |= 4;
	if( HaveFloor( row-1, col ) )
		Direc |= 2;
	if( HaveFloor( row+1, col ) )
		Direc |= 8;
	return Direc;
}

int PathAPP::DrawFloor( int row, int col, MapPtr Ima )
{
	CalPosition( row, col );
	BYTE PPB = PixelPerBlock - 1;
	WORD FloorChk = 0;

	if( *(Map+PosByte) & PosBit )
	{
		FloorChk = 1;
		if( Final[FinalSelected].Row == row && Final[FinalSelected].Col == col )
			memcpy( Ima, FinalIma, SizeFloor );
		else
			memcpy( Ima, FloorIma, SizeFloor );
		// Check Top Map
		if( !HaveFloor( row-1, col ) )
			ImaLine( Ima, 0, 0, PPB, 0, WallColor );
		// Check Bottom Map
		if( !HaveFloor( row+1, col ) )
			ImaLine( Ima, 0, PPB, PPB, PPB, WallColor );
		// Check Left Map
		if( !HaveFloor( row, col-1 ) )
			ImaLine( Ima, 0, 0, 0, PPB, WallColor );
		// Check Right Map
		if( !HaveFloor( row, col+1 ) )
			ImaLine( Ima, PPB, 0, PPB, PPB, WallColor );
	}
	else
	{
		int far *IntPtr = (int far *)Ima;
		*IntPtr++ = PPB;
		*IntPtr = PPB;
		ImaBar( Ima, 0, 0, PPB, PPB, BLACK );
		if( !row )
		{
			ImaLine( Ima, 0, 0, PPB, 0, WallColor );
			FloorChk = 1;
		}
		else if( row == MaxRow-1 )
		{
			ImaLine( Ima, 0, PPB, PPB, PPB, WallColor );
			FloorChk = 1;
		}
		if( !col )
		{
			ImaLine( Ima, 0, 0, 0, PPB, WallColor );
			FloorChk = 1;
		}
		else if( col == MaxCol-1 )
		{
			ImaLine( Ima, PPB, 0, PPB, PPB, WallColor );
			FloorChk = 1;
		}
	}
	return FloorChk;
}


int PathAPP::HaveFinal( int row, int col )
{
	for( register int i = 0 ; i < MaxFinal ; i++ )
	{
		if( Final[i].Row == row && Final[i].Col == col )
			return 1;
	}
	return 0;
}

void PathAPP::LineWall( int row, int col )
{
	int x = PixelPerBlock*col; // StartX = StartY = 0
	int y = PixelPerBlock*row;
	BYTE PPB = PixelPerBlock - 1;

	setcolor( WallColor );
	// Check Top Map
	if( !HaveFloor( row-1, col ) )
		line( x, y, x+PPB, y );
	// Check Bottom Map
	if( !HaveFloor( row+1, col ) )
		line( x, y+PPB, x+PPB, y+PPB );
	// Check Left Map
	if( !HaveFloor( row, col-1 ) )
		line( x, y, x, y+PPB );
	// Check Right Map
	if( !HaveFloor( row, col+1 ) )
		line( x+PPB, y, x+PPB, y+PPB );
}

#if defined(GAME)
void PathAPP::DrawFloor( int row, int col )
{
//	int x = StartX + PixelPerBlock*col;
//	int y = StartY + PixelPerBlock*row;
	int x = PixelPerBlock*col; // StartX = StartY = 0
	int y = PixelPerBlock*row;

	CalPosition( row, col );
	BYTE PPB = PixelPerBlock - 1;

	if( *(Map+PosByte) & PosBit )
	{
		if( Final[FinalSelected].Row == row && Final[FinalSelected].Col == col )
			putimage( x, y, FinalIma, COPY_PUT );
		else
			putimage( x, y, FloorIma, COPY_PUT );
		setcolor( WallColor );
		// Check Top Map
		if( !HaveFloor( row-1, col ) )
			line( x, y, x+PPB, y );
		// Check Bottom Map
		if( !HaveFloor( row+1, col ) )
			line( x, y+PPB, x+PPB, y+PPB );
		// Check Left Map
		if( !HaveFloor( row, col-1 ) )
			line( x, y, x, y+PPB );
		// Check Right Map
		if( !HaveFloor( row, col+1 ) )
			line( x+PPB, y, x+PPB, y+PPB );
	}
	else
	{
		setfillstyle( SOLID_FILL, BLACK );
		bar( x, y, x+PPB, y+PPB );
		setcolor( WallColor );
		if( !row )
			line( x, y, x+PPB, y );
		else if( row == MaxRow-1 )
			line( x, y+PPB, x+PPB, y+PPB );
		if( !col )
			line( x, y, x, y+PPB );
		else if( col == MaxCol-1 )
			line( x+PPB, y, x+PPB, y+PPB );
	}
}
#endif

#if defined(MAP)
void PathAPP::DrawFloor( int row, int col )
{
//	int x = StartX + PixelPerBlock*col;
//	int y = StartY + PixelPerBlock*row;
	int x = PixelPerBlock*col; // StartX = StartY = 0
	int y = PixelPerBlock*row;


	if( HaveFinal( row, col ) )
	{
		putimage( x, y, FinalIma, COPY_PUT );
		LineWall( row, col );
		return;
	}

	CalPosition( row, col );
	BYTE PPB = PixelPerBlock - 1;


	if( *(Map+PosByte) & PosBit )
	{
		putimage( x, y, FloorIma, COPY_PUT );
		LineWall( row, col );
	}
	else
	{
		setfillstyle( SOLID_FILL, BLACK );
		bar( x, y, x+PPB, y+PPB );
		setcolor( WallColor );
		if( !row )
			line( x, y, x+PPB, y );
		else if( row == MaxRow-1 )
			line( x, y+PPB, x+PPB, y+PPB );
		if( !col )
			line( x, y, x, y+PPB );
		else if( col == MaxCol-1 )
			line( x+PPB, y, x+PPB, y+PPB );
	}
}
#endif


void PathAPP::UpdateFloor( int row, int col )
{
	DrawFloor( row, col );
	if( row > 0 )
		DrawFloor( row-1, col );
	if( row < MaxRow-1 )
		DrawFloor( row+1, col );
	if( col > 0 )
		DrawFloor( row, col-1 );
	if( col < MaxCol-1 )
		DrawFloor( row, col+1 );
}

WORD PathAPP::HaveFloor( int row, int col )
{
	if( row < 0 )
		return 0;
	if( col < 0 )
		return 0;
	if( row >= MaxRow )
		return 0;
	if( col >= MaxCol )
		return 0;
	CalPosition( row, col );
	if( *(Map+PosByte) & PosBit )
		return 1;
	return 0;
}
                     
void PathAPP::CalPosition( int row, int col )
{
	PosByte = row * ((MaxCol-1)/8 + 1);
	PosByte += col/8;
	PosBit = 0X80;
	PosBit >>= col%8;
}


typedef struct {
	int StartX, StartY, MaxRow, MaxCol;
	BYTE PixelPerBlock, WallColor;
	int FinalR, FinalC;
	size_t SizeMap, SizeFloor;
	} PathFile;

typedef struct {
	char PickText[40];
	BYTE Version;
	char Designer[100];
	BYTE UnusedA[30];   // keep PathFILE size = 256
	int MaxRow, MaxCol;
	BYTE PixelPerBlock, WallColor;
	size_t SizeMap, SizeFloor;
	BYTE MaxFinal;
	struct {
		int Row, Col;
		} Final[MAX_FINAL];
	BYTE UnusedB[30];  // keep PathFILE size = 256
	BYTE GameXORChk, PickTextChk;
	BYTE SumChk, XORChk;
	} PathFILE;

int PathAPP::WriteFile( char *filename )
{
	FILE *fout = fopen( filename, "wb" );
	if( !fout )
		return 0;
	PathFILE PF;
	memset( (void*)&PF, 0, 256 );

	strcpy( PF.PickText, "Pick map file." );
	PF.PickText[14] = (char)0X0D;
	PF.PickText[15] = (char)0X0A;
	PF.PickText[16] = (char)0X1A;
	PF.Version = 0X01;
	strcpy( PF.Designer, (char *)Designer );

	PF.MaxRow = MaxRow;
	PF.MaxCol = MaxCol;
	PF.PixelPerBlock = PixelPerBlock;
	PF.WallColor = WallColor;

	PF.SizeMap = SizeMap;
	PF.SizeFloor = SizeFloor;

	PF.MaxFinal = MaxFinal;
	memcpy( (void *)&PF.Final, (void *)&Final, 4*MaxFinal );

	BYTE *ChkPtr = (BYTE *)(&PF);
	ChkPtr += 40;
	for( register int i = 0 ; i < 182 ; i++ )
	{
		PF.SumChk += *ChkPtr;
		PF.XORChk ^= *ChkPtr++;
	}

	if( !fwrite((void *)&PF,sizeof(PathFILE),1,fout) )
	{
		fclose( fout );
		return 0;
	}
	if( !fwrite((void *)Map,PF.SizeMap,1,fout) )
	{
		fclose( fout );
		return 0;
	}
	if( !fwrite((void *)FloorIma,PF.SizeFloor,1,fout) )
	{
		fclose( fout );
		return 0;
	}

	fclose( fout );
	return 1;
}


int PathAPP::ReadFile( char *filename )
{
	FILE *fin = fopen( filename, "rb" );
	if( !fin )
		return 0;

	struct ffblk FileBlk;
	if( findfirst(filename,&FileBlk,0) )
	{
		fclose( fin );
		return 0;
	}

	if( FileBlk.ff_fsize != 320 )
	{
		BYTE SumChk = 0;
		BYTE XORChk = 0;

		PathFILE PF;

		fread( (void *)&PF, sizeof(PathFILE), 1, fin );
		BYTE *ChkPtr = (BYTE *)(&PF);
		ChkPtr += 40;
		// Check map file
		for( register int i = 0 ; i < 182 ; i++ )
		{
			SumChk += *ChkPtr;
			XORChk ^= *ChkPtr++;
		}

		if( SumChk != PF.SumChk || XORChk != PF.XORChk ||
			PF.Version != 0X01 )
		{
			fclose( fin );
			return 0;
		}

		strcpy( (char *)Designer, PF.Designer );
		MaxRow = PF.MaxRow;
		MaxCol = PF.MaxCol;
		PixelPerBlock = PF.PixelPerBlock;
		WallColor = PF.WallColor;
		SizeFloor = PF.SizeFloor;
		SizeMap = PF.SizeMap;

		MaxFinal = PF.MaxFinal;
		memcpy( (void *)&Final, (void *)&PF.Final, 4*MaxFinal );

		StartX = StartY = 0;

		Map = (MapPtr) farmalloc( PF.SizeMap );
		if( !Map )
		{
			fclose( fin );
			return 0;
		}
		FloorIma = (MapPtr) farmalloc( PF.SizeFloor );
		if( !FloorIma )
		{
			fclose( fin );
			return 0;
		}
		if( !fread((void *)Map,PF.SizeMap,1,fin) )
		{
			fclose( fin );
			return 0;
		}
		if( !fread((void *)FloorIma,PF.SizeFloor,1,fin ) )
		{
			fclose( fin );
			return 0;
		}

		fclose( fin );
		return 1;
	}

	PathFile PFile;
	fread( (void *)&PFile, sizeof(PathFile), 1, fin );

	strcpy( (char *)Designer, "" );
	StartX = StartY = 0;
	MaxRow = PFile.MaxRow;
	MaxCol = PFile.MaxCol;
	PixelPerBlock = PFile.PixelPerBlock;
	WallColor = PFile.WallColor;
	MaxFinal = 1;
	Final[0].Row = PFile.FinalR;
	Final[0].Col = PFile.FinalC;

	SizeFloor = PFile.SizeFloor;
	SizeMap = PFile.SizeMap;

	Map = (MapPtr) farmalloc( PFile.SizeMap );
	if( !Map )
	{
		fclose( fin );
		return 0;
	}
	FloorIma = (MapPtr) farmalloc( PFile.SizeFloor );
	if( !FloorIma )
	{
		fclose( fin );
		return 0;
	}
	fread( (void *)Map, PFile.SizeMap, 1, fin );
	fread( (void *)FloorIma, PFile.SizeFloor, 1, fin );
	fclose( fin );
	return 1;
}

PathAPP::~PathAPP( void )
{
	if( FloorIma )
		farfree( FloorIma );
	if( Map )
		farfree( Map );
	if( FinalIma )
		farfree( FinalIma );
}
